from dataclasses import dataclass, field
from functools import cached_property
from itertools import product
from typing import Any, Iterable, Sequence, TypeVar

from bloqade.geometry.dialects import grid
from kirin import types
from kirin.dialects import ilist

NumX = TypeVar("NumX")
NumY = TypeVar("NumY")


@dataclass(eq=False)
class FilledGrid(grid.Grid[NumX, NumY]):
    x_spacing: tuple[float, ...] = field(init=False)
    y_spacing: tuple[float, ...] = field(init=False)
    x_init: float | None = field(init=False)
    y_init: float | None = field(init=False)

    parent: grid.Grid[NumX, NumY]
    vacancies: frozenset[tuple[int, int]]

    def __post_init__(self):
        self.x_spacing = self.parent.x_spacing
        self.y_spacing = self.parent.y_spacing
        self.x_init = self.parent.x_init
        self.y_init = self.parent.y_init

        self.type = types.Generic(
            FilledGrid,
            types.Literal(len(self.x_spacing) + 1),
            types.Literal(len(self.y_spacing) + 1),
        )

    def __hash__(self):
        return hash((self.parent, self.vacancies))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, FilledGrid)
            and self.parent == other.parent
            and self.vacancies == other.vacancies
        )

    def is_equal(self, other: Any) -> bool:
        return self == other

    @cached_property
    def positions(self) -> ilist.IList[tuple[float, float], Any]:
        positions = tuple(
            (x, y)
            for (ix, x), (iy, y) in product(
                enumerate(self.x_positions), enumerate(self.y_positions)
            )
            if (ix, iy) not in self.vacancies
        )

        return ilist.IList(positions)

    @classmethod
    def fill(
        cls, grid_obj: grid.Grid[NumX, NumY], filled: Sequence[tuple[int, int]]
    ) -> "FilledGrid[NumX, NumY]":
        num_x, num_y = grid_obj.shape

        if isinstance(grid_obj, FilledGrid):
            vacancies = grid_obj.vacancies
        else:
            vacancies = frozenset(product(range(num_x), range(num_y)))

        vacancies = vacancies - frozenset(filled)

        return cls(parent=grid_obj, vacancies=vacancies)

    @classmethod
    def vacate(
        cls, grid_obj: grid.Grid[NumX, NumY], vacancies: Iterable[tuple[int, int]]
    ) -> "FilledGrid[NumX, NumY]":

        if isinstance(grid_obj, FilledGrid):
            input_vacancies = grid_obj.vacancies
        else:
            input_vacancies = frozenset()

        input_vacancies = input_vacancies.union(vacancies)

        return cls(parent=grid_obj, vacancies=input_vacancies)

    def get_view(  # type: ignore
        self, x_indices: ilist.IList[int, Any], y_indices: ilist.IList[int, Any]
    ):
        remapping_x = {ix: i for i, ix in enumerate(x_indices)}
        remapping_y = {iy: i for i, iy in enumerate(y_indices)}
        return FilledGrid(
            parent=self.parent.get_view(x_indices, y_indices),
            vacancies=frozenset(
                (remapping_x[x], remapping_y[y])
                for x, y in self.vacancies
                if x in remapping_x and y in remapping_y
            ),
        )

    def shift(self, x_shift: float, y_shift: float):
        return FilledGrid(
            parent=self.parent.shift(x_shift, y_shift),
            vacancies=self.vacancies,
        )

    def scale(self, x_scale: float, y_scale: float):
        return FilledGrid(
            parent=self.parent.scale(x_scale, y_scale),
            vacancies=self.vacancies,
        )

    def repeat(self, x_times: int, y_times: int, x_gap: float, y_gap: float):
        new_parent = self.parent.repeat(x_times, y_times, x_gap, y_gap)
        x_dim, y_dim = self.shape
        vacancies = frozenset(
            (x + x_dim * i, y + y_dim * j)
            for i, j, (x, y) in product(range(x_times), range(y_times), self.vacancies)
        )
        return FilledGrid.vacate(new_parent, vacancies)


FilledGridType = types.Generic(FilledGrid, types.TypeVar("NumX"), types.TypeVar("NumY"))



from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import gate, init, measure, spec
from bloqade.shuttle.analysis.runtime import RuntimeAnalysis
from bloqade.shuttle.prelude import move
from bloqade.shuttle.stdlib.waypoints import move_by_waypoints


def test_simple_true():
    @move
    def main():
        i = 2
        init.fill([spec.get_static_trap(zone_id="test")])
        return i

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_simple_false():
    @move
    def main():
        return 1

    assert not RuntimeAnalysis(move).has_quantum_runtime(main)


def test_if_1():
    @move
    def main():
        i = 0
        if i % 2 == 0:
            return measure.measure((spec.get_static_trap(zone_id="test"),))

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_if_2():
    @move
    def main():
        i = 0
        if i % 2 == 0:
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_if_3():
    @move
    def main(cond: bool):
        i = 0
        start = spec.get_static_trap(zone_id="test")
        end = grid.shift(start, 1, 0)
        if cond:
            move_by_waypoints(ilist.IList([start, end]))
            i = i + 1
        else:
            return 1

        return i

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_if_4():
    @move
    def main(cond: bool):
        i = 0
        if cond:
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))
            return i + 2
        else:
            return i + 1

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_if_5():
    @move
    def main(cond: bool):
        i = 0
        if cond:
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))
            return i + 2
        else:
            i = i + 1

        return i

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_loop_1():
    @move
    def main():
        i = 0
        for i in range(1, 9, 2):
            if i % 2 == 0:
                gate.top_hat_cz(spec.get_static_trap(zone_id="test"))

    assert not RuntimeAnalysis(move).has_quantum_runtime(main)


def test_loop_2():
    @move
    def main():
        i = 0
        for i in range(0, 9, 2):
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_loop_3():
    @move
    def main():
        i = 0
        for i in range(0, 9, 2):
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))
            return i

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_subroutine_1():
    @move
    def subroutine():
        gate.top_hat_cz(spec.get_static_trap(zone_id="test"))

    @move
    def main():
        i = 0
        subroutine()
        return i

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_subroutine_2():
    @move
    def subroutine(a: int):
        return a + 1

    @move
    def main(i: int):
        return subroutine(i)

    assert not RuntimeAnalysis(move).has_quantum_runtime(main)


def test_lambda_1():
    @move
    def main():
        def inner():
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))

        inner()

    assert RuntimeAnalysis(move).has_quantum_runtime(main)


def test_lambda_2():
    @move
    def main(a: int):
        def inner(i: int):
            return i + 1

        return inner(a)

    assert not RuntimeAnalysis(move).has_quantum_runtime(main)


def test_lambda_3():
    @move
    def main(a: int):
        def inner(i: int):
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))
            return i + 1

        return inner

    assert not RuntimeAnalysis(move).has_quantum_runtime(main)


def test_lambda_4():
    @move
    def subroutine(a: int):
        def inner(i: int):
            gate.top_hat_cz(spec.get_static_trap(zone_id="test"))
            return a + i

        return inner

    @move
    def main(i: int):
        f = subroutine(i)
        return f(i)

    assert RuntimeAnalysis(move).has_quantum_runtime(main)



from bloqade.geometry.dialects import grid
from bloqade.types import QubitType
from kirin import ir, lowering, types
from kirin.decl import info, statement
from kirin.dialects import ilist

from ._dialect import dialect
from .types import AtomType

NumAtoms = types.TypeVar("NumAtoms")


@statement(dialect=dialect)
class New(ir.Statement):
    name = "new"

    traits = frozenset({lowering.FromPythonCall()})

    zone: ir.SSAValue = info.argument(grid.GridType[types.Any, types.Any])
    qubits: ir.SSAValue = info.argument(ilist.IListType[QubitType, NumAtoms])
    result: ir.ResultValue = info.result(ilist.IListType[AtomType, NumAtoms])


@statement(dialect=dialect)
class Move(ir.Statement):
    name = "move"

    traits = frozenset({lowering.FromPythonCall(), ir.Pure()})

    zone: ir.SSAValue = info.argument(grid.GridType[types.Any, types.Any])
    atoms: ir.SSAValue = info.argument(ilist.IListType[AtomType, NumAtoms])
    result: ir.ResultValue = info.result(ilist.IListType[AtomType, NumAtoms])


@statement(dialect=dialect)
class MoveNextTo(ir.Statement):
    name = "move_next_to"

    traits = frozenset({lowering.FromPythonCall(), ir.Pure()})

    zone: ir.SSAValue = info.argument(grid.GridType[types.Any, types.Any])
    ctrls: ir.SSAValue = info.argument(ilist.IListType[AtomType, NumAtoms])
    qargs: ir.SSAValue = info.argument(ilist.IListType[AtomType, NumAtoms])

    def __init__(self, zone: ir.SSAValue, ctrls: ir.SSAValue, qargs: ir.SSAValue):
        NumAtoms = (
            ctrls.type.vars[1]
            if isinstance(ctrls.type, types.Generic)
            else types.TypeVar("NumAtoms")
        )
        super().__init__(
            args=(zone, ctrls, qargs),
            args_slice={"zone": 0, "ctrls": 1, "qargs": 2},
            result_types=[
                ilist.IListType[AtomType, NumAtoms],
                ilist.IListType[AtomType, NumAtoms],
            ],
        )


@statement(dialect=dialect)
class ResetPosition(ir.Statement):
    name = "reset_position"

    traits = frozenset({lowering.FromPythonCall()})

    atoms: ir.SSAValue = info.argument(ilist.IListType[AtomType, NumAtoms])
    qubits: ir.SSAValue = info.argument(ilist.IListType[QubitType, NumAtoms])


@statement(dialect=dialect)
class Measure(ir.Statement):
    name = "measure"

    traits = frozenset({lowering.FromPythonCall()})

    atoms: ir.SSAValue = info.argument(ilist.IListType[AtomType, NumAtoms])
    qubits: ir.SSAValue = info.argument(ilist.IListType[QubitType, NumAtoms])
    result: ir.ResultValue = info.result(ilist.IListType[types.Int, NumAtoms])



import pytest
from kirin import interp, ir
from kirin.dialects import ilist

from bloqade.shuttle.dialects import action, schedule
from bloqade.shuttle.prelude import move, tweezer


@tweezer
def test_move():
    action.turn_on(action.ALL, action.ALL)
    action.turn_off(action.ALL, action.ALL)


class TestScheduleInterpreter:

    def init_interpreter(self):
        interpreter = interp.Interpreter(move)
        interpreter.initialize()

        return interpreter

    def run_stmt(self, stmt_type: type[ir.Statement], *values):
        interpreter = self.init_interpreter()
        ssa_values = tuple(ir.TestValue() for _ in values)
        new_stmt = stmt_type(*ssa_values)
        return interpreter.run_stmt(new_stmt, values)

    def test_new_device_function(self):

        x_tones = ilist.IList([1, 2, 3])
        y_tones = ilist.IList([4, 5, 6])

        result = self.run_stmt(
            schedule.NewDeviceFunction,
            test_move,
            x_tones,
            y_tones,
        )

        assert result == (
            schedule.DeviceFunction(
                move_fn=test_move, x_tones=x_tones, y_tones=y_tones
            ),
        )

    def test_reverse(self):
        x_tones = ilist.IList([1, 2, 3])
        y_tones = ilist.IList([4, 5, 6])

        device_task = schedule.DeviceFunction(
            move_fn=test_move,
            x_tones=x_tones,
            y_tones=y_tones,
        )
        reversed_task = schedule.ReverseDeviceFunction(device_task=device_task)

        result = self.run_stmt(schedule.Reverse, device_task)
        assert result == (reversed_task,)

        result = self.run_stmt(schedule.Reverse, reversed_task)
        assert result == (device_task,)

        with pytest.raises(interp.InterpreterError):
            self.run_stmt(schedule.Reverse, "invalid_task")



from bloqade.geometry.dialects import grid

from bloqade.shuttle import spec
from bloqade.shuttle.passes import inject_spec
from bloqade.shuttle.prelude import move


def get_spec(slm_grid: grid.Grid) -> spec.ArchSpec:
    return spec.ArchSpec(
        layout=spec.Layout(
            {"slm": slm_grid},
            fillable=set(["slm"]),
            has_cz=set(["slm"]),
            has_local=set(["slm"]),
        )
    )


def test_inject_spec():
    slm_grid = grid.Grid.from_positions([1, 2, 3], [4, 5, 6])
    test_spec = get_spec(slm_grid)

    @move(arch_spec=test_spec)
    def test():
        return spec.get_static_trap(zone_id="slm")

    assert (
        test() == slm_grid
    ), "The injected static trap should match the expected grid."


def test_inject_spac_callgraph():
    slm_grid = grid.Grid.from_positions([1, 2, 3], [4, 5, 6])
    test_spec = get_spec(slm_grid)

    @move
    def subroutine(depth: int):
        slm = spec.get_static_trap(zone_id="slm")

        if depth >= 1:

            def lambda_func():
                return slm

            return lambda_func

        return subroutine(depth + 1)

    @move
    def test():
        res = subroutine(0)
        return res()

    inject_spec.InjectSpecsPass(move, arch_spec=test_spec, fold=False)(test)

    test.verify_type()

    assert (
        test() == slm_grid
    ), "The injected static trap should match the expected grid."


def test_inject_constants():
    test_spec = spec.ArchSpec(
        int_constants={"my_int": 42},
        float_constants={"my_float": 3.14},
    )

    @move(arch_spec=test_spec)
    def test():
        return spec.get_int_constant(constant_id="my_int"), spec.get_float_constant(
            constant_id="my_float"
        )

    assert test() == (
        42,
        3.14,
    ), "The injected constants should match the expected values."



from kirin.interp import (
    Frame,
    Interpreter,
    InterpreterError,
    MethodTable,
    impl,
)

from bloqade.shuttle.codegen import TraceInterpreter, reverse_path
from bloqade.shuttle.dialects import schedule
from bloqade.shuttle.dialects.path import dialect, stmts, types


@dialect.register
class PathInterpreter(MethodTable):

    @impl(stmts.Gen)
    def gen(self, interp: Interpreter, frame: Frame, stmt: stmts.Gen):
        if stmt.arch_spec is None:
            raise InterpreterError("Missing architecture specification")

        device_task = frame.get(stmt.device_task)
        if isinstance(device_task, schedule.DeviceFunction):
            reverse = False
        elif isinstance(device_task, schedule.ReverseDeviceFunction):
            device_task = device_task.device_task
            reverse = True
        else:
            raise InterpreterError("Invalid device task type")

        inputs = frame.get_values(stmt.inputs)
        kwargs = stmt.kwargs
        args = interp.permute_values(device_task.move_fn.arg_names, inputs, kwargs)
        path = TraceInterpreter(stmt.arch_spec).run_trace(device_task.move_fn, args, {})

        if reverse:
            path = reverse_path(path)

        return (
            types.Path(
                x_tones=device_task.x_tones,
                y_tones=device_task.y_tones,
                path=path,
            ),
        )



from dataclasses import dataclass, field

from bloqade.geometry.dialects import grid
from kirin import interp, ir
from kirin.analysis.forward import Forward, ForwardFrame

from bloqade.shuttle.arch import ArchSpec

from .lattice import SpecZone, UnknownZone, Zone


@dataclass
class ZoneAnalysis(Forward[Zone]):
    """
    Analysis for zones in the Qourier context.
    This analysis extends the Forward analysis to work with Zone types.
    """

    keys = ["zone.analysis"]
    lattice = Zone
    arch_spec: ArchSpec = field(default_factory=ArchSpec)

    def get_grid_lattice(self, zone: grid.Grid) -> Zone:
        for zone_id, zone_obj in self.arch_spec.layout.static_traps.items():
            if zone is zone_obj:  # must be the object defined in the spec to match
                return SpecZone(zone_id)

        return UnknownZone()

    def eval_stmt_fallback(
        self, frame: ForwardFrame[Zone], stmt: ir.Statement
    ) -> tuple[Zone, ...] | interp.SpecialValue[Zone]:
        return tuple(
            (
                self.lattice.top()
                if result.type.is_subseteq(grid.GridType)
                else self.lattice.bottom()
            )
            for result in stmt.results
        )

    def run_method(self, method: ir.Method, args: tuple[Zone, ...]):
        # NOTE: we do not support dynamic calls here, thus no need to propagate method object
        return self.run_callable(method.code, (self.lattice.bottom(),) + args)



# %% [markdown]
# # Programming the move layout for logical magic state distillation

# %% [markdown]
# In this example, we program the move layout employed in the demonstration of logical
# magic state distillation for a distance-5 color code, as presented in QuEra’s recent
# publication: [**“Experimental demonstration of logical magic state distillation”,
# Nature, 2025**](https://arxiv.org/pdf/2506.14169).
# This example focuses specifically on the atom moves and two-qubit gate operations
# between atom pairs. For clarity, we omit the single-qubit gates used in the full
# circuit. The circuits executed in the paper were carefully optimized for the
# architectural features of QuEra’s neutral atom hardware. In particular, they leverage
# extensive parallelization of atom shuttling operations and simultaneous execution of
# two-qubit gates.
#
# We will demonstrate that `bloqade.shuttle` provides a concise and efficient interface
# for expressing such parallelized atom movements.
#
# As a first step, we import the relevant modules of `bloqade.shuttle` and `kirin`.

# %%
from typing import Any, TypeVar

import matplotlib
from bloqade.geometry.dialects import grid
from kirin.dialects import ilist

from bloqade.shuttle import gate, init, spec
from bloqade.shuttle.prelude import move
from bloqade.shuttle.stdlib.spec import single_zone_spec
from bloqade.shuttle.stdlib.waypoints import move_by_waypoints
from bloqade.shuttle.visualizer import PathVisualizer

# %% [markdown]
# The primary subroutines used to implement atom transport are the `move` dialect and
# the `move_by_waypoints` function from the standard library. Programming atom movement
# in shuttle is centered around atoms arranged on two-dimensional grids. This reflects
# the capabilities of QuEra’s Gemini-class hardware, which utilizes pairs of
# acousto-optic deflectors (AODs) to manipulate atom positions independently in the
# x- and y-directions.
#
# Grid-based layouts are chosen to optimize both the accessibility and
# parallelizability of atom moves, given AOD constraints such as simultaneous pickup of
# atoms aligned along columns and rows, as well as the need to prevent inter-atomic
# collisions during transport. The shuttle framework encodes AOD constraints directly
# into the semantics of move kernel objects. Attempting to compile or execute a move
# sequence that violates these constraints will result in a runtime or compile-time
# error, ensuring physical feasibility of the programmed layout on hardware

# %% [markdown]
# As a first step, we define three global parameters that govern the geometry and
# dynamics of the atom array:
#
# - `grid_spacing`: Specifies the lattice spacing of the initial two-dimensional grid.
# This grid corresponds to the static trap array defined by the spatial light modulator
# (SLM), which generates the optical tweezers used to hold the atoms.
#
# - `entangling_pair_dist`: Sets the target distance between pairs of atoms that are
# intended to undergo a two-qubit entangling gate. This distance must be sufficiently
# small to enable a strong Rydberg-mediated interaction, which is essential for
# high-fidelity entanglement. A typical value used in experiments is 2 μm.
#
# - `path_shift_dist`: During atom transport, it is often necessary to temporarily
# displace atoms off-grid to avoid collisions when crossing rows or columns. This
# parameter defines the offset applied to atoms during their movement, enabling
# collision-free shuttling through the array.

# %%
# define global geometry choices
grid_spacing = 10.0  # spacing of the grid traps

entangling_pair_dist = 2.0  # distance between atom pairs getting entangled

path_shift_dist = (
    3.0  # distance that is used to shift the atom path out of the columns or rows
)

# %% [markdown]
# # Basic move pattern

# %% [markdown]
# In this section, we program the precise atom movement pattern used in the experimental
# demonstration of logical magic state distillation. Due to the high degree of
# connectivity required by the circuit, atom transport is restricted to full-row and
# full-column movements, maximizing parallelism while respecting hardware constraints.
#
# We begin by specifying the initial atom grid configuration. This example focuses on a
# single entangling zone layout, where each logical qubit is encoded along a single row
# of atoms. For a distance-5 color code, each logical qubit comprises 17 physical
# qubits. The experimental implementation involved five such logical qubits.
#
# This geometry is conveniently defined using the `single_zone_spec` function, where:
#
# - `num_x` specifies the number of columns (i.e., physical qubits per logical qubit),
# - `num_y` specifies the number of rows (i.e., logical qubits), and
# - `spacing` defines the uniform distance between adjacent rows and columns.
#
# By construction, the grid is assumed to be equidistant in both spatial dimensions.

# %%
arch_spec = single_zone_spec(num_x=17, num_y=5, spacing=grid_spacing)

# %% [markdown]
# We define a set of helper kernel functions to enable coordinated movement of atoms
# along rows and columns of the grid. The core routine, `entangle_cols`, operates on two
# `IList` arguments -- `ctrls` and `qargs`. The `ctrls` list specifies the row indices
# of control qubits to be moved, while `qargs` indicates the row indices of the target
# qubits with which entanglement is to be performed.
#
# To register the function as a valid movement kernel for compilation to
# `bloqade.shuttle` IR, we decorate it with `@move`. This ensures compatibility with
# `kirin`’s intermediate representation and hardware-aware compilation pipeline.
#
# The structure of the entangling routine comprises the following key steps:
#
# 1. Define initial trap sites: Specify the locations of static SLM-defined traps where
# all qubits are initially held.
#
# 2. Construct subgrids: Define subgrids corresponding to the atoms to be picked up
# (`ctrls`), and subgrids defining the positions of the target qubits (`qargs`) near
# which the control atoms will be temporarily positioned to enable interaction.
#
# 3. Specify waypoint trajectories: Construct ordered sequences of intermediate
# positions (waypoints) to guide atom transport between subgrids.
#
# 4. Invoke `move_by_waypoints`: Use this function to concatenate the defined path
# segments, producing the full motion trajectory between initial and final subgrid
# configurations.
#
# It is important to note that, in this layout, atoms always return to their original
# static trap positions following each entangling operation. This symmetry enables easy
# definition of the reverse path by simply inverting the sequence of waypoints.
#
# The `move_by_waypoints` function includes Boolean flags that indicate whether atoms
# should be picked up from or dropped into static traps at each stage. In the context
# of this circuit, we maintain atoms within the AODs during gate execution and only
# return them to their SLM-defined traps afterward. This is a deliberate design choice
# and not a general constraint. More sophisticated pickup/dropoff sequences can be
# constructed using multiple defined trap grids. However, care must be taken: dropping
# an atom at a site lacking an active trap will result in atom loss.


# %%
@move
def entangle_cols(ctrls: ilist.IList[int, Any], qargs: ilist.IList[int, Any]):
    """Function to entangle columns of atoms."""

    # fill the defined zone specification with traps
    zone = spec.get_static_trap(zone_id="traps")
    # get the shape of the trap array
    traps_shape = grid.shape(zone)
    # generate an ilist from 0 to the number of rows in the trap array
    all_rows = ilist.range(traps_shape[1])

    # subgrid of control qubits that are picked up
    src = grid.sub_grid(zone, ctrls, all_rows)
    # subgrid of target qubits that are entangled with the control qubits
    dst = grid.sub_grid(zone, qargs, all_rows)

    # define the moves
    # shifting the src grid down along the y-axis
    first_waypoint = grid.shift(src, 0.0, -path_shift_dist)

    # shift the to the respective x-positions of the target
    # qubits and add an offset in the x-direction
    second_waypoint = grid.shift(dst, -entangling_pair_dist, -path_shift_dist)

    # shift the dst grid back up into the original y-position
    # to form atom pairs that will get a gate
    third_waypoint = grid.shift(dst, -entangling_pair_dist, 0.0)

    # combine the waypoints into an ilist
    waypoints = ilist.IList([src, first_waypoint, second_waypoint, third_waypoint])
    # the revese waypoints defining the reverse path back to the original positions
    reverse_waypoints = ilist.IList(
        [third_waypoint, second_waypoint, first_waypoint, src]
    )

    # move the qubits along the waypoints, True means that the atoms are picked up
    # and False means that they are not dropped at the end of the move
    move_by_waypoints(waypoints, True, False)
    # apply the entangling gate to the atoms that are now in the right positions
    # (paired up)
    gate.top_hat_cz(zone)
    # move the atoms back to their original positions, False means that the atoms
    # are not picked up (since they are still in the AOD) and True means that they
    # are dropped back into their static trap site at the end of the move
    move_by_waypoints(reverse_waypoints, False, True)


# %% [markdown]
# In an analogous manner, we can now define parallel transport operations for atoms
# along rows of the grid.


# %%
@move
def entangle_rows(ctrls: ilist.IList[int, Any], qargs: ilist.IList[int, Any]):
    """Function to entangle rows of atoms."""

    # get the positions of the static traps in the entangling zone
    zone = spec.get_static_trap(zone_id="traps")
    # get the shape of the trap array
    traps_shape = grid.shape(zone)
    # generate an ilist from 0 to the number of columns in the trap array
    all_cols = ilist.range(traps_shape[0])

    src = grid.sub_grid(zone, all_cols, ctrls)
    dst = grid.sub_grid(zone, all_cols, qargs)

    # define the moves
    # shift the src grid to the right along the x-axis
    first_waypoint = grid.shift(src, entangling_pair_dist, 0.0)
    # move the src grid to the y-positions of the target qubits
    second_waypoint = grid.shift(dst, entangling_pair_dist, 0.0)

    # combine the waypoints into an ilist
    waypoints = ilist.IList([src, first_waypoint, second_waypoint])
    # the revese waypoints defining the reverse path back to the original positions
    reverse_waypoints = ilist.IList([second_waypoint, first_waypoint, src])

    # move the qubits along the waypoints, True means that the atoms are picked up
    # and False means that they are not dropped at the end of the move
    move_by_waypoints(waypoints, True, False)
    # apply the entangling gate to the atoms that are now in the right positions
    # (paired up)
    gate.top_hat_cz(zone)
    # move the atoms back to their original positions, False means that the atoms
    # are not picked up (since they are still in the AOD) and True means that they
    # are dropped back into their static trap site at the end of the move
    move_by_waypoints(reverse_waypoints, False, True)


# %% [markdown]
# Using these helper functions, we can now specify the complete move pattern in just a
# few lines of code by providing the control and target atom indices to the
# `entangle_cols` and `entangle_rows` functions.
#
# As before, we define this as a kernel function, annotated with the `@move` decorator
# to indicate compatibility with the bloqade.shuttle IR. This kernel is wrapped inside
# a closure and returned as a first-class move kernel function.


# %%
def make_main(entangle_cols, entangle_rows):
    """Helper function to create the main move kernel for logical magic state
    distillation.

    Args
        entangle_cols: Function to entangle columns of atoms.
        entangle_rows: Function to entangle rows of atoms.

    Returns
        main: The main move kernel function that defines the entire move pattern.

    """

    @move
    def main():
        """Main move kernel function that defines the entire move pattern for the
        logical magic state distillation experiment.

        """

        init.fill([spec.get_static_trap(zone_id="traps")])

        # encode logical qubits by entangling pairs of atoms in columns
        entangle_cols([1, 10, 12, 13], [3, 7, 14, 16])
        entangle_cols([4, 8, 11, 15], [7, 10, 14, 16])
        entangle_cols([2, 8, 9, 10, 14], [4, 6, 7, 13, 16])
        entangle_cols([0, 3, 5, 10, 11], [2, 6, 8, 12, 13])
        entangle_cols([0, 2, 4, 6, 8, 12], [1, 3, 5, 7, 9, 15])

        # logical qubit operations by moving entire rows of atoms
        entangle_rows([0, 2], [1, 3])
        entangle_rows([1, 3], [2, 4])
        entangle_rows([3, 4], [0, 1])

    return main


ker = make_main(entangle_cols, entangle_rows)

# %% [markdown]
# We can verify the correctness of the programmed move layout using the `PathVisualizer`
# utility provided by `bloqade.shuttle`. This tool displays the atom trajectories
# between defined waypoints and enables stepwise inspection of the full movement
# sequence via the `Continue` button. Red flashes are used to indicate the application
# of two-qubit gate pulses at the entangling zone during the sequence.

# %%
matplotlib.use("TkAgg")  # requirement for PathVisualizer

PathVisualizer(ker.dialects, arch_spec=arch_spec).run(ker, ())

# %% [markdown]
# # Further refining the move pattern

# %% [markdown]
# We now take a further step by introducing a slight optimization to the move pattern.
# In the previous example, control qubits were moved in both leftward and rightward
# directions, but ultimately all were positioned to the left of their respective target
# qubits. This introduces unnecessary displacement for half of the control atoms. To
# minimize total movement, we can instead position control atoms to the nearest side of
# the target qubits, i.e., those moving rightward are placed to the left, and those
# moving leftward are placed to the right of the targets.
#
# To implement this, we define an additional helper function that computes the nearest
# feasible final position for each control qubit, given the target layout and movement
# direction.

# %%
N = TypeVar("N")


@move
def get_final_positions(
    src: ilist.IList[float, N], dst: ilist.IList[float, N], offset: float
):
    """Helper function to compute the nearest final positions for entanglement."""

    assert len(src) == len(
        dst
    ), "Source and destination lists must be of the same length."

    def get_last_pos(i: int) -> float:
        assert src[i] != dst[i], "Source and destination positions must not be equal."
        if src[i] < dst[i]:
            return dst[i] - offset
        elif src[i] > dst[i]:
            return dst[i] + offset
        else:
            return dst[i]

    return ilist.map(get_last_pos, ilist.range(len(src)))


# %% [markdown]
# Using this helper function, we can now construct new waypoint sequences that
# incorporate the nearest final positions for the control qubits.


# %%
@move
def entangle_cols_low_dist(ctrls: ilist.IList[int, Any], qargs: ilist.IList[int, Any]):
    """Helper function to entangle columns of atoms on a grid in a single entangling
    zone with optimized final positions (nearest location).

    """

    zone = spec.get_static_trap(zone_id="traps")
    traps_shape = grid.shape(zone)
    all_rows = ilist.range(traps_shape[1])

    src = grid.sub_grid(zone, ctrls, all_rows)
    dst = grid.sub_grid(zone, qargs, all_rows)

    first_waypoint = grid.shift(src, 0.0, -path_shift_dist)

    dst_x = grid.get_xpos(dst)
    src_x = grid.get_xpos(src)

    last_x = get_final_positions(
        src_x, dst_x, entangling_pair_dist
    )  # get the nearest final positions for the control qubits

    second_pos = grid.from_positions(last_x, grid.get_ypos(first_waypoint))
    last_pos = grid.from_positions(last_x, grid.get_ypos(dst))

    waypoints = ilist.IList([src, first_waypoint, second_pos, last_pos])
    reverse_waypoints = ilist.IList([last_pos, second_pos, first_waypoint, src])

    move_by_waypoints(waypoints, True, False)
    gate.top_hat_cz(zone)
    move_by_waypoints(reverse_waypoints, False, True)


# %% [markdown]
# We can now define a new move kernel that implements the optimized column-wise
# transport pattern using the updated waypoint assignments.

# %%
ker = make_main(entangle_cols_low_dist, entangle_rows)

# %% [markdown]
# Finally, we can once again visualize the optimized move pattern using the
# PathVisualizer to inspect the resulting trajectories and validate the
# updated layout.

# %%
matplotlib.use("TkAgg")  # requirement for PathVisualizer

PathVisualizer(ker.dialects, arch_spec=arch_spec).run(ker, ())



from .gate import GateVisualizerMethods as GateVisualizerMethods
from .init import InitVisualizerMethods as InitVisualizerMethods
from .measure import MeasureVisualizerMethods as MeasureVisualizerMethods
from .path import PathVisualizerMethods as PathVisualizerMethods



from dataclasses import dataclass, field

from kirin import ir, rewrite
from kirin.dialects.py import Constant
from kirin.ir.nodes.stmt import Statement
from kirin.passes import Fold, Pass
from kirin.passes.callgraph import CallGraphPass
from kirin.rewrite.abc import RewriteResult, RewriteRule

from bloqade.shuttle.arch import ArchSpec
from bloqade.shuttle.dialects import path, spec


@dataclass
class InjectSpecRule(RewriteRule):
    arch_spec: ArchSpec

    def rewrite_Statement(self, node: Statement) -> RewriteResult:
        if isinstance(node, path.Gen) and node.arch_spec is None:
            node.arch_spec = self.arch_spec
            return RewriteResult(has_done_something=True)
        elif (
            isinstance(node, spec.GetStaticTrap)
            and (zone_id := node.zone_id) in self.arch_spec.layout.static_traps
        ):
            node.replace_by(Constant(self.arch_spec.layout.static_traps[zone_id]))

            return RewriteResult(has_done_something=True)
        elif (
            isinstance(node, spec.GetIntConstant)
            and node.constant_id in self.arch_spec.int_constants
        ):
            node.replace_by(Constant(self.arch_spec.int_constants[node.constant_id]))
            return RewriteResult(has_done_something=True)
        elif (
            isinstance(node, spec.GetFloatConstant)
            and node.constant_id in self.arch_spec.float_constants
        ):
            node.replace_by(Constant(self.arch_spec.float_constants[node.constant_id]))
            return RewriteResult(has_done_something=True)

        return RewriteResult()


@dataclass
class InjectSpecsPass(Pass):
    arch_spec: ArchSpec
    fold: bool = True
    fold_pass: Fold = field(init=False)

    def __post_init__(self):
        self.fold_pass = Fold(self.dialects, no_raise=self.no_raise)

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:
        # since we're rewriting `mt` inplace we should make sure it is on the visited list
        # so that recursive calls are handed correctly
        rule = rewrite.Walk(InjectSpecRule(self.arch_spec))
        result = CallGraphPass(self.dialects, rule, no_raise=self.no_raise).unsafe_run(
            mt
        )
        if self.fold:
            result = self.fold_pass(mt).join(result)

        return result



from . import runtime as runtime
from ._dialect import dialect as dialect
from ._interface import measure as measure
from .stmts import Measure as Measure, New as New
from .types import (
    MeasurementArray as MeasurementArray,
    MeasurementArrayType as MeasurementArrayType,
    MeasurementResult as MeasurementResult,
    MeasurementResultType as MeasurementResultType,
)



from bloqade.geometry.dialects.grid.types import Grid, SubGrid
from kirin.analysis.forward import ForwardFrame
from kirin.dialects import py
from kirin.interp import MethodTable, impl

from .. import analysis, lattice


@py.indexing.dialect.register(key="zone.analysis")
class GetIndexImpl(MethodTable):
    @impl(py.indexing.GetItem)
    def get_item(
        self,
        interp: analysis.ZoneAnalysis,
        frame: ForwardFrame[lattice.Zone],
        stmt: py.indexing.GetItem,
    ):
        obj = frame.get(stmt.obj)

        if isinstance(obj, lattice.InvalidZone):
            return (lattice.InvalidZone(),)

        if isinstance(obj, lattice.Zone):
            return (lattice.GetItemOfZone(obj, frame.get(stmt.index)),)

        return (lattice.NotZone(),)


@py.constant.dialect.register(key="zone.analysis")
class ConstantImpl(MethodTable):
    @impl(py.Constant)
    def constant(
        self,
        interp: analysis.ZoneAnalysis,
        frame: ForwardFrame[lattice.Zone],
        stmt: py.Constant,
    ):
        zone = stmt.value.unwrap()

        if isinstance(zone, SubGrid):
            # handle cases where ir has been folded
            zone_lattice = interp.get_grid_lattice(zone.parent)
            if isinstance(zone_lattice, lattice.InvalidZone):
                return (lattice.InvalidZone(),)

            return (
                lattice.GetSubGridOfZone(
                    zone=interp.get_grid_lattice(zone.parent),
                    x_indices=lattice.Zone.bottom(),
                    y_indices=lattice.Zone.bottom(),
                ),
            )
        elif isinstance(zone, Grid):
            return (interp.get_grid_lattice(zone),)

        return (lattice.NotZone(),)



from dataclasses import dataclass
from typing import TypeVar

from kirin import interp

from bloqade.shuttle.dialects.path import Parallel, Path, Play, dialect
from bloqade.shuttle.visualizer.interp import PathVisualizer
from bloqade.shuttle.visualizer.renderers import RendererInterface


@dataclass
class ParallelRuntime:
    paths: tuple[Path, ...]

    def __post_init__(self):
        if any(not isinstance(p, Path) for p in self.paths):
            raise interp.InterpreterError("path is not canonicalized")


@dialect.register(key="path.visualizer")
class PathVisualizerMethods(interp.MethodTable):

    Renderer = TypeVar("Renderer", bound=RendererInterface)

    @interp.impl(Play)
    def play(self, _interp: PathVisualizer[Renderer], frame: interp.Frame, stmt: Play):

        path = frame.get(stmt.path)
        # path is generated by "main" interpreter so
        # we need to render the path before showing it
        if isinstance(path, Path):

            _interp.renderer.render_path(path)
        elif isinstance(path, ParallelRuntime):
            for p in path.paths:
                _interp.renderer.render_path(p)
        else:
            raise interp.InterpreterError(
                f"Expected a Path or tuple of Paths, got {path}"
            )

        return ()

    @interp.impl(Parallel)
    def parallel(
        self, _interp: "PathVisualizer[Renderer]", frame: interp.Frame, stmt: Parallel
    ):
        return (ParallelRuntime(frame.get_values(stmt.paths)),)



from typing import Any

from bloqade.geometry.dialects import grid
from kirin.dialects import ilist
from kirin.lowering import wraps as _wraps

from .stmts import Fill


@_wraps(Fill)
def fill(
    locations: ilist.IList[grid.Grid[Any, Any], Any] | list[grid.Grid[Any, Any]],
):
    """Fill the given locations with the given value.

    Args:
        locations (ilist.IList[grid.Grid[Any,Any], Any]): The locations to fill.
            Note that these locations must be valid static trap locations.

    """
    ...



from ._dialect import dialect as dialect
from ._interface import (
    ALL as ALL,
    move as move,
    set_loc as set_loc,
    turn_off as turn_off,
    turn_on as turn_on,
)
from .stmts import (
    IntensityStatement as IntensityStatement,
    Move as Move,
    Set as Set,
    TurnOff as TurnOff,
    TurnOffXSlice as TurnOffXSlice,
    TurnOffXY as TurnOffXY,
    TurnOffXYSlice as TurnOffXYSlice,
    TurnOffYSlice as TurnOffYSlice,
    TurnOn as TurnOn,
    TurnOnXSlice as TurnOnXSlice,
    TurnOnXY as TurnOnXY,
    TurnOnXYSlice as TurnOnXYSlice,
    TurnOnYSlice as TurnOnYSlice,
    TweezerFunction as TweezerFunction,
)
